#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

""" This module holds the code related to particle data. """

# Standard library imports


class Particle:
    """
    This class exists to hold the data on particle populations, creation and loss.
    """

    def __init__(self, creation_data, loss_data, total_data, particle_type):
        self.attributes = []
        self.particle_type = particle_type
        for item in creation_data:
            name = ' '.join(item.split()) + ' creation'
            name = name.title()
            self.__dict__[name] = creation_data[item]
            self.attributes.append(name)
        for item in loss_data:
            name = ' '.join(item.split()) + ' loss'
            name = name.title()
            self.__dict__[name] = loss_data[item]
            self.attributes.append(name)
        for item in total_data:
            name = ' '.join(item.split()).title()
            self.__dict__[name] = total_data[item]
            self.attributes.append(name)

    def __repr__(self):
        """ Print the attributes of the particle to the terminal."""
        print()
        print(f"Printing {self.particle_type} Data...".upper())
        for item in self.attributes:
            print(f"{item}: {self.__dict__[item]}")
        print()




#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

""" This module creates the HTML output for the SCALE data """

# Standard library imports
import datetime
try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources
# Third party imports
from jinja2 import Template
# local imports
try:
    from .. import static
except ValueError:
    import static


def get_css():
    """Get the css to be used in the html output

    Returns:
        inline_css (str): The whole static css file as a single string
    """
    inline_css = pkg_resources.read_text(static, 'style.css')
    return inline_css


def get_html(case):
    """Get the html output from the jinja template

    Args:
        case (EddySCALECase): The EddySCALECase object for this Eddy run

    Returns:
        html (str): the html output as a single string
    """
    inline_css = get_css()
    html_template = pkg_resources.read_text(static, 'SCALE_template.html')
    template = Template(html_template)
    # render template as a unicode string
    html = template.render(
        filename=case.filepath,
        case_name=case.name,
        # logo=f"{os.getcwd()}/static/logo.png",
        inline_css=inline_css,
        rundate=case.rundate,
        runtime=case.runtime,
        date=datetime.datetime.now().strftime("%Y/%m/%d"),
        time=datetime.datetime.now().strftime("%H:%M:%S"),
        scaling_factor=case.scaling_factor,
        tally_list=case.tally_list,
        mixture_list=case.mixture_list,
        input_deck=case.scale_input,
        )
    return html



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

"""Eddy is a programme to convert the text-based output files from MCNP and SCALE to
a user-friendly HTML format.

Copyright (C) 2020  Cerberus Nuclear Ltd

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


This module is the entry point for Eddy, the MCNP and SCALE to HTML output converter.
This script takes an output file and scaling factor, determines whether it is an MCNP or SCALE case,
determines whether it is a criticality or shielding case
and creates the relevant EddyMCNPCase or EddySCALECase object,
then calls the relevant HTML writer
"""

# Imports from standard library
import re
import os.path
from markupsafe import escape
# Local imports
from .scale import scale_html_writer
from .scale.eddy_scale_case import EddySCALECase
from .mcnp import mcnp_html_writer
from .mcnp.eddy_mcnp_case import EddyMCNPCase


class NotAcceptedFileTypeError(Exception):
    """Raised when the file selected is not recognised as a SCALE or MCNP output"""


def main(filename, scaling_factor=1):
    """Entry point to Eddy. Requires filename, optional scaling factor as arguments.
    Gets the output data and determines whether it is a crit case.
    Determine whether it is an MCNP or SCALE case;
    calls the relevant converter.

    Args:
        filename (str): the file path (including the name) of the output file
        scaling_factor (float): A number by which any tally results will be multiplied
    """

    # check file exists
    assert os.path.isfile(filename), "That output file does not exist in that location."

    # Check MCNP or SCALE, criticality or shielding case, get output data
    output_data, code, crit_case = get_args(filename)

    if code == 'SCALE':
        case = EddySCALECase(
            filepath=filename,
            file=output_data,
            scaling_factor=scaling_factor,
        )
        html = scale_html_writer.get_html(case)
    elif code == 'MCNP':
        case = EddyMCNPCase(
            filepath=filename,
            scaling_factor=scaling_factor,
            file=output_data,
            crit_case=crit_case,
        )
        html = mcnp_html_writer.get_html(case)
    else:
        raise NotAcceptedFileTypeError("This file doesn't seem to be an MCNP or SCALE output?")

    output_file, extension = os.path.splitext(filename)
    output_file += '.html'
    write_output(output_file, html)
    print(f"Eddy run complete, {output_file} created.\n")


def get_args(filename):
    """
    Determine if this is a crit case (for MCNP).
    Determine if this is a SCALE CASE
    Read in the output data from the mcnp output file.

    Args:
        filename (str): the file path (including the name) of the output file

    Returns:
        output_data (list): The contents of the output file
        crit_case (bool): True if kcode case, otherwise False
        code (str): 'SCALE' or 'MCNP'
    """

    # get contents of file
    output_data = read_file(filename)

    if 'SCALE' in output_data[2]:
        code = 'SCALE'
    elif 'Code Name &amp; Version = MCNP' in output_data[0]:
        code = 'MCNP'
    else:
        raise NotAcceptedFileTypeError("This file doesn't seem to be an MCNP or SCALE output?")

    # Check if shielding or crit case
    if code == 'MCNP':
        crit_case = check_if_crit(output_data)
    else:
        crit_case = False

    print(f"Output file selected: {filename}")
    return output_data, code, crit_case


def read_file(filename):
    """Read in the output file.

    Args:
        filename (str): The name (and location) of the .out file

    Returns:
        data (list): The lines from the output file
    """
    with open(filename, 'r') as file:
        data = file.readlines()
    data = sanitize_list(data)
    return data


def sanitize_list(text):
    """Replace any html control characters in any string in the
    list with the appropriate html codes to stop
    them from being interpreted as html tags.

    Args:
        text [list]: a list of strings

    Returns:
        list: The sanitized version of the list
    """
    sanitized_text = []
    for line in text:
        line = str(escape(line))
        sanitized_text.append(line)
    return sanitized_text


def check_if_crit(output_data):
    """Read through output data to check if case is crit or shielding

    Args:
        output_data (list): The text of the output file

    Returns:
        True if kcode found or False if not.

    """
    PATTERN_crit = re.compile(r"\s+\d+-\s{7}(kcode).*")
    for line in output_data:
        if re.match(PATTERN_crit, line):
            return True
    return False


def write_output(file, html):
    """Write the HTML to the desired file

    Args:
        file (str): The filepath of the html file to be written
        html (str): The contents to be written to the html file
    """
    with open(file, 'w') as f:
        f.write(html)





#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

""" This module contains all the code relating to SCALE tallies """


class Tally:
    """Each SCALE tally is represented by a Tally object."""
    def __init__(self, data):
        self.tally_number = data[0].split()[3][:-1]
        self.type = f"{data[0].split()[1]} {data[0].split()[2]}"
        try:
            self.title = data[0].split('.')[1]
        except IndexError:
            self.title = None
        self.particle = data[0].split()[0]
        self.response = float(data[6].split()[2])
        if self.response:
            self.rel_uncertainty = float(data[6].split()[4])
        else:
            self.rel_uncertainty = 0.0
        statistical_checks = data[6][69:80].split()
        self.passes = 0
        for check in statistical_checks:
            if check == 'X':
                self.passes += 1
        self.checks = {}
        for check, result in enumerate(statistical_checks, start=1):
            self.checks[check] = ('Pass' if result == 'X' else 'Fail')


    def scale_results(self, scaling_factor):
        """Scale the results of each tally by the project-level scaling factor

        Args:
            scaling_factor (float): A number to multiply results by
            """
        self.response *= scaling_factor


# END OF TALLY CLASS


def get_tally_data(output_data):
    """Find the parts of the SCALE output with the tally data

    Args:
        output_data (list): The SCALE output data

    Returns:
        tally_data (list): The lines from the SCALE output containing the tally data
    """
    for n, line in enumerate(output_data):
        if "Final Tally Results Summary" in line:
            for m, other_line in enumerate(output_data[n+1:], start=n+1):
                if "Total Monaco cpu time for this problem" in other_line:
                    tally_data = output_data[n:m]
                    return tally_data
    # If the function has not returned the data, there is a problem
    raise Exception("Tally data incomplete.")


def create_tallies(tally_data):
    """Create Tally objects using the output data

    Args:
        tally_data (list): The lines from the SCALE output concerning tallies

    Returns:
        None, but creates Tally objects
    """
    for n, line in enumerate(tally_data):
        if "Region Tally" in line or "Point Detector" in line:
            tally_info = tally_data[n:n+7]
            Tally(tally_info)



#!/usr/bin/python3
# Peter Evans
# Cerberus Nuclear Ltd
# November 2019

"""
    This module holds all the code related to cells, including the Cells class,
    and the functions which find the cell data in the mcnp output, extract the cell
    section of that data, and send it to the Cell class to create the Cell objects.

    It also contains the code to find the neutron, photon and electron population data,
    which is assigned to each cell object as they are created.
"""
# imports from standard library
import re
# local imports


class Cell:
    """ This class holds the attributes and methods for Cell objects."""
    # pylint: disable=too-many-instance-attributes
    # MCNP cells have a lot of attributes

    def __init__(self, **info):
        self.cell_number = info['cell_number']
        self.material_number = info['material_number']
        self.atom_density = info['atom_density']
        self.gram_density = info['gram_density']
        self.volume = float(info['volume'])
        self.mass = float(info['mass'])
        self.neutron_importance = info['neutron_importance']
        self.photon_importance = info['photon_importance']
        self.electron_importance = info['electron_importance']

    def assign_populations(self, neutron_pop=None, photon_pop=None, electron_pop=None):
        """Assign neutron, photon and electron population data to each cell.

        Args:
            neutron_pop (list): The neutron population section of the MCNP output
            photon_pop (list): The photon population section of the MCNP output
            electron_pop (list): The electron population section of the MCNP output

        Returns:
            None. The data is assigned to the cell objects as various new attributes.
        """
        # Assign neutron population data to cell
        if neutron_pop:
            for line in neutron_pop[6:-2]:
                line = line.split()
                num = line[1]
                if num == self.cell_number:
                    self.neutron_tracks_entering = line[2]
                    self.neutron_population = line[3]
                    self.neutron_collisions = line[4]
        # Assign photon population data to cell
        if photon_pop:
            for line in photon_pop[6:-2]:
                line = line.split()
                num = line[1]
                if num == self.cell_number:
                    self.photon_tracks_entering = line[2]
                    self.photon_population = line[3]
                    self.photon_collisions = line[4]
        # Assign electron population data to cell
        if electron_pop:
            for line in electron_pop[6:-2]:
                line = line.split()
                num = line[1]
                if num == self.cell_number:
                    self.electron_tracks_entering = line[2]
                    self.electron_population = line[3]
                    self.electron_collisions = line[4]

    def __repr__(self):
        """Print a description of the cell object to the terminal (for debugging purposes)

        Args:
            self, the cell object

        Returns:
            None
        """
        description = (f"Cell number: {self.cell_number}"
                       f"Cell material: {self.material_number}"
                       f"Cell density: {self.gram_density} g/cm3"
                       f"Cell volume: {self.volume} cm3")
        try:
            description += f"\nNeutron Importance: {self.neutron_importance}"
        except AttributeError:
            pass
        try:
            description += f"\nPhoton Importance: {self.photon_importance}"
        except AttributeError:
            pass
        try:
            description += f"\nElectron Importance: {self.electron_importance}"
        except AttributeError:
            pass
        return description




#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

"""
    This module holds all the code relating to tallies.
"""

# standard library imports
import re


# local imports:


class Tally:
    """Each mcnp tally is represented by a Tally object containing all the available data about that tally.
    All tallies should be members of one of the subclass of this Tally class.
    """

    def __init__(self, data):
        # for item in data: print(item)
        self.data = data  # can be removed once this method is complete
        self.tally_number = data[0].split()[1]
        self.nps = data[0].split()[4]
        self.f_type = 'F' + data[1].split()[2]
        self.tally_type = data[1].split()[3:]
        self.tally_type = ' '.join(self.tally_type)
        self.particles = data[2].split()[1]
        self.dose_functions = self.get_dose_functions()
        self.results = self.get_results()
        self.statistical_checks = self.get_statistical_checks()
        self.passes = self.get_passes()

    def get_dose_functions(self):
        """Gets the dose function, if any, that the results of this tally are multiplied by

        Returns:
            Either a tuple of the DE and DF functions, if they exist, otherwise a string
            stating that this tally is not multiplied by a dose function.

        """
        if "this tally is modified by dose function" in self.data[3]:
            return self.data[3].split()[7], self.data[3].split()[9][:-1]
        else:
            return "This tally is not modified by any dose function"

    def get_results(self):
        """This is a placeholder method; each subclass of Tally should have its own 'get_results' method"""
        raise NotImplementedError(f"The {self.__class__} subclass should have its own get_results() method")

    def get_statistical_checks(self):
        """Get the statistical check results from the mcnp output file

        Returns:
            statistical_checks, a dictionary with the results of the statistical checks for this tally.
        """
        start_PATTERN = re.compile(r"\s+results of 10 statistical checks.+")
        value = None
        pass_fail = None
        for n, line in enumerate(self.data):
            if start_PATTERN.match(line):
                value = self.data[n + 6].split()
                pass_fail = self.data[n + 7].split()
                break
        if value:  # if statistical checks have been found
            statistical_checks = {
                'mean_behaviour': {'value': value[1], 'pass': pass_fail[1]},
                'rel_err_value': {'value': value[2], 'pass': pass_fail[2]},
                'rel_err_decrease': {'value': value[3], 'pass': pass_fail[3]},
                'rel_err_dec_rate': {'value': value[4], 'pass': pass_fail[4]},
                'VoV_value': {'value': value[5], 'pass': pass_fail[5]},
                'VoV_decrease': {'value': value[6], 'pass': pass_fail[6]},
                'VoV_dec_rate': {'value': value[7], 'pass': pass_fail[7]},
                'FoM_value': {'value': value[8], 'pass': pass_fail[8]},
                'FoM_behaviour': {'value': value[9], 'pass': pass_fail[9]},
                'pdf_slope': {'value': value[10], 'pass': pass_fail[10]},
            }
        else:
            # print(f"Tally {self.tally_number} does not have statistical checks")
            statistical_checks = {
                'mean_behaviour': {'value': r"none", 'pass': "no"},
                'rel_err_value': {'value': r"none", 'pass': "no"},
                'rel_err_decrease': {'value': r"none", 'pass': "no"},
                'rel_err_dec_rate': {'value': r"none", 'pass': "no"},
                'VoV_value': {'value': r"none", 'pass': "no"},
                'VoV_decrease': {'value': r"none", 'pass': "no"},
                'VoV_dec_rate': {'value': r"none", 'pass': "no"},
                'FoM_value': {'value': r"none", 'pass': "no"},
                'FoM_behaviour': {'value': r"none", 'pass': "no"},
                'pdf_slope': {'value': r"none", 'pass': "no"},
            }
        return statistical_checks

    def get_passes(self):
        """Determine how many statistical checks were passed

        Returns:
            Passes (int): the number of checks passed (out of ten)
        """
        passes = 0
        if self.statistical_checks:
            for check in self.statistical_checks:
                if self.statistical_checks[check]['pass'] == 'yes':
                    passes += 1
        return passes

    def scale_result(self, scaling_factor):
        """Each subclass of Tally should have its own normalise_data method"""
        print("This Tally class appears not to have its own 'normalise_data() method.")
        raise NotImplementedError(f"The {self.__class__} subclass should have its own normalise_data method")

    def describe_object(self):
        """ Print a description of the Tally object to the terminal."""
        print()
        print(f"Tally {self.tally_number}")
        print(f"Tally type: {self.f_type}, {self.tally_type}")
        print(f"Particles: {self.particles}")
        print(f"Number of particles: {self.nps}")
        print(f"This tally is modified by dose functions {self.dose_functions[0]} and {self.dose_functions[1]}")
        print(f"Tally Results: ")
        for region in self.results:
            print(f"    {region['region']}:     Tally Result = {region['result']}      Variance = {region['variance']}")
        print()
        if 'statistical_checks' in dir(self):
            print("Statistical Check      Value      Pass?")
            for check in self.statistical_checks:
                print(
                    f"{check + ':' : <21}  {self.statistical_checks[check]['value'] : <11} {self.statistical_checks[check]['pass'] : <3}")
            print(f"{self.passes} out of 10 statistical checks passed.")
            print()
        else:
            print("This tally does not have statistical checks.")


class F2Tally(Tally):
    def __init__(self, data):
        super().__init__(data)

    def get_results(self):
        """
            Gets the tally results from the mcnp output file
            Args: self: the object, data: the mcnp output tally section
            Returns: regions: a dictionary holding the results for that tally
        """
        data = self.data
        results = []
        for num, line in enumerate(data):
            if "surface " in line:
                results.append({
                    "surface": line.strip().capitalize(),
                    "result": float(data[num + 1].split()[0]),
                    "variance": float(data[num + 1].split()[1])
                })
        return results

    def scale_result(self, scaling_factor):
        """Applies scaling factor to results
            Args:
                scaling_factor (float): a number by which the region results are multiplied
            Returns: none, but modifies self.results
        """
        for surface in self.results:
            surface['result'] *= scaling_factor


class F4Tally(Tally):
    def __init__(self, data):
        super().__init__(data)

    def get_results(self):
        """Get the tally results from the mcnp output file

        Returns:
            list: A list of dictionaries, each corresponding to a cell
                            in this tally, with entries for region, result and variance.
        """
        data = self.data
        results = []
        for num, line in enumerate(data):
            if (" cell " in line) or (" surface  " in line):
                results.append({
                    "region": line.strip().capitalize(),
                    "result": float(data[num + 1].split()[0]),
                    "variance": float(data[num + 1].split()[1])
                })
        return results

    def scale_result(self, scaling_factor):
        """Applies scaling factor to results
            Args:
                scaling_factor (float): a number by which the region results are multiplied
            Returns: none, but modifies self.results
        """
        for region in self.results:
            region['result'] *= scaling_factor


class F5Tally(Tally):
    def __init__(self, data):
        super().__init__(data)

    def get_results(self):
        """
            Gets the tally results from the mcnp output file
            Args: self: the object, data: the mcnp output tally section
            Returns: regions: a dictionary holding the results for that tally
        """
        data = self.data
        results = []
        for num, line in enumerate(data):
            if "detector located at" in line and "uncollided" not in data[num + 1]:
                results.append({
                    "x": float(line[28:40]),
                    "y": float(line[40:52]),
                    "z": float(line[52:64]),
                    "result": float(data[num + 1].split()[0]),
                    "variance": float(data[num + 1].split()[1]),
                })
        return results

    def scale_result(self, scaling_factor):
        """Applies scaling factor to results
            Args:
                scaling_factor (float): a number by which the region results are multiplied
            Returns: none, but modifies self.results
        """
        for detector in self.results:
            detector['result'] *= scaling_factor


class F6Tally(Tally):
    def __init__(self, data):
        super().__init__(data)
        if self.f_type == "F6+":
            self.particles = "Collision Heating"

    def get_results(self):
        data = self.data
        """Get the tally results from the MCNP output file
        Args:
            data (list): the section of the MCNP output file for this tally
        Returns:
            results (dict): A dictionary of dictionaries of dictionaries, each corresponding to a cell
                            in this tally, with entries for region, mass, result 
                            and variance.
        """
        results = {}
        for num, line in enumerate(data):
            if 'masses' in line:
                masses_start = num
                break
        for num, line in enumerate(data[masses_start:], start=masses_start):
            if 'cell  ' in line:
                masses_end = num - 1
                break
        mass_data = data[masses_start:masses_end]
        for num, line in enumerate(mass_data):
            if "cell" in line:
                for n, mass in enumerate(mass_data[num + 1].split()):
                    cell_no = mass_data[num].split()[n + 1]
                    results[cell_no] = {'region': f"Cell {cell_no}", 'mass': float(mass)}

        PATTERN_f6_cell = re.compile(r'^\s+cell\s+\d+')
        for num, line in enumerate(data):
            if 'cell ' in line:
                # if PATTERN_f6_cell.match(line):
                cell_no = line.split()[1]
                results[cell_no]["result"] = float(data[num + 1].split()[0])
                results[cell_no]["variance"] = float(data[num + 1].split()[1])

        return results

    def scale_result(self, scaling_factor):
        """Applies scaling factor to results
            Args:
                scaling_factor (float): a number by which the region results are multiplied
            Returns: none, but modifies self.results
        """
        for region in self.results:
            self.results[region]['result'] *= scaling_factor

############################################################
#  End of Tally class                                      #
############################################################



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

"""This module contains the EddyMCNPCase class, which holds all the data relevant to a particular MCNP run.

"""

import re
from .cells import Cell
from .particles import Particle
from .tallies import F2Tally, F4Tally, F5Tally, F6Tally


class EddyMCNPCase:
    """ An EddyCase object holds all the essential data for a single Eddy run,
    and its __init__ method is responsible for calling the other classes that
    Eddy uses.
    """
    def __init__(self, filepath, scaling_factor, file, crit_case=False):
        """ Most of the work done by eddy is handled within the __init__ function;
        it calls all the instance methods that get information from the mcnp output file,
        and puts them into attributes of the EddyCase object.

        Args:
            filepath (str): The path to the mcnp output file
            scaling_factor (float): A number to multiply the tally results by
            file (list): The contents of the mcnp output file
            crit_case (bool): True if crit case, otherwise false
        """

        self.filepath = filepath
        self.name = filepath.replace('\\', '/').split('/')[-1]
        self.scaling_factor = scaling_factor
        self.file = file
        self.crit_case = crit_case

        # start parsing output:
        self.rundate, self.runtime = self.get_date_time()
        self.ctme, self.nps = self.get_runtime()
        self.mcnp_input = self.get_input()
        self.parameters = self.get_parameters()
        self.lost_particles = self.check_lost_particles()
        self.fatal_errors = self.get_fatal_errors()
        self.warnings = self.get_warnings()
        self.comments = self.get_comments()
        self.duplicate_surfaces = self.get_duplicate_surfaces()
        if self.crit_case is True:
            self.k_effective = self.get_k_eff()
            self.cycles = self.get_active_cycles()
        else:
            self.k_effective = None
            self.cycles = None

        # Cells
        self.cell_data = self.get_cell_data()
        self.neutron_populations = self.get_particle_populations('neutron')
        self.photon_populations = self.get_particle_populations('photon')
        self.electron_populations = self.get_particle_populations('electron')
        self.cell_list = self.create_cells()
        for cell in self.cell_list:
            cell.assign_populations(
                self.neutron_populations,
                self.photon_populations,
                self.electron_populations,
            )

        # Particles
        self.particle_list = []
        for particle_type in ['neutron', 'photon', 'electron']:
            particle_data = self.find_mcnp_particle_data(particle_type)
            if particle_data:
                particle = self.create_particle(particle_type, particle_data)
                self.particle_list.append(particle)

        # Tallies
        # initialise empty variables
        self.tally_list = []   # list of all tallies
        self.f_types = []
        self.F2_tallies = {'neutrons': [], 'photons': [], 'electrons': []}
        self.F4_tallies = {'neutrons': [], 'photons': [], 'electrons': []}
        self.F5_tallies = {'neutrons': [], 'photons': [], 'electrons': []}
        self.F6_tallies = {'neutrons': [], 'photons': [], 'electrons': [], 'Collision Heating': []}
        # get tallies
        if self.crit_case is False:
            # TODO: sort this monstrosity of a function call out
            self.tally_list, self.f_types, self.F2_tallies, self.F4_tallies, self.F5_tallies, self.F6_tallies = self.get_tallies()
        # apply scaling factor
            if self.scaling_factor != 1:
                for tally in self.tally_list:
                    tally.scale_result(self.scaling_factor)

    def __repr__(self):
        return (f"This Eddy MCNP case considers file {self.name}\n"
                f"from {self.filepath}\n"
                f"The scaling factor is {self.scaling_factor}\n"
                f"Criticality case: {self.crit_case}\n")

    def get_date_time(self):
        """Get the date and time that the mcnp case was run

        Returns:
            tuple: The rundate and runtime, as a tuple of 2 strings
        """
        for line in self.file:
            if line.startswith('1mcnp'):
                time = line.split()[5]
                date = line.split()[4]
                date = date.split('/')
                d = date[1]
                m = date[0]
                y = '20' + date[2]
                # date_time = {'date': f"{y}/{m}/{d}", 'time': time}
                rundate = f"{y}/{m}/{d}"
                runtime = time
                return rundate, runtime

    def get_runtime(self):
        """Get the time or number of particles the MCNP case was run for

        Returns:
            tuple:
                str:
                    The number of minutes the code ran for, or None if not found.
                str:
                    nps The number of particles run, or None if not found.

        """
        ctme = None
        nps = None
        pattern_nps = re.compile(r'^\s{6,}\d+-\s{7}(nps|NPS)\s+.+')
        pattern_ctme = re.compile(r'^\s{6,}\d+-\s{7}(ctme|CTME)\s+.+')
        for line in self.file:
            if pattern_ctme.match(line):
                ctme = line.split()[2]
            if pattern_nps.match(line):
                nps = line.split()[2]
        return ctme, nps

    def get_input(self):
        """Extract the MCNP input deck from the MCNP output file

        Returns:
            list: The MCNP input deck, as a list of strings
        """
        pattern_input = re.compile(r'^\s{6,}\d+-\s{7}(.*)')
        mcnp_input = []
        for line in self.file:
            match = pattern_input.match(line)
            if match:
                mcnp_input.append(match.group(1).strip('\n'))
        return mcnp_input

    def get_parameters(self):
        """Get any parameters used in the input file.
        NOTE: This is only useful if MCNP is run using the Cerberus
        package CYCLONE; non-cyclone users can ignore this function.

        Returns:
            dict: A dictionary containing all the parameters used by this case,
                    and their values
        """
        variables = {}
        for n, line in enumerate(self.mcnp_input):
            if "USING THE FOLLOWING VARIABLES" in line:
                for m, other_line in enumerate(self.mcnp_input[n+1:], start=n+1):
                    if '=' not in other_line:
                        break
                    variable = other_line.split()[1]
                    value = other_line.split()[3]
                    variables[variable] = float(value)
                break
        return variables

    def check_lost_particles(self):
        """Check whether the run was terminated because 10 or more particles got lost

        Returns:
            bool: True if run terminated due to lost particles, otherwise False

        """
        for line in self.file:
            if "run terminated because" in line and "particles got lost" in line:
                return True
        return False

    def get_fatal_errors(self):
        """Find the warnings in the MCNP output data

        Returns:
            list: A list of the fatal error messages
        """
        fatal_errors = []
        for line in self.file:
            if "fatal error." in line:
                message = line[14:].strip().capitalize()
                fatal_errors.append(message)
        return fatal_errors

    def get_warnings(self):
        """Find the warnings in the MCNP output data

        Returns:
            list: A list of the warning messages
        """
        warnings = []
        PATTERN_warnings = re.compile(r'warning')
        for line in self.file:
            if PATTERN_warnings.search(line):
                if "warning message so far" not in line and "warning messages so far" not in line:
                    warning = line[10:].strip().capitalize()
                    if warning not in warnings:
                        warnings.append(warning)
        return warnings

    def get_comments(self):
        """Find the comments in the MCNP output file.

        Returns:
            list: A list of the comments
        """
        PATTERN_comments = re.compile(r'comment\.\s+[A-Za-z0-9].+')    # Ignores blank comment lines
        comments = []
        for line in self.file:
            if PATTERN_comments.search(line):
                comment = line[10:].strip().capitalize()
                comments.append(comment)
        return comments

    def get_duplicate_surfaces(self):
        """Find all the duplicate surface messages in the mcnp output

        Returns:
            list: A list of the duplicate surfaces
        """
        duplicate_surfaces = []
        PATTERN_duplicates = re.compile(r'\ssurface\s+\d+.+and surface.+are the same.+')
        for line in self.file:
            if PATTERN_duplicates.match(line):
                if line.strip().capitalize() not in duplicate_surfaces:
                    duplicate_surfaces.append(line.strip().capitalize())
        return duplicate_surfaces

    def get_k_eff(self):
        """Search the output data for the section concerning k-effective, and create a single dictionary holding the data

        Returns:
            dict: A dictionary with the k-effective values for the first half, second half and total run
        """
        PATTERN_k_eff = re.compile(r'^\s*problem\s+keff.+')
        k_eff = {}
        for num, line in enumerate(self.file):
            if PATTERN_k_eff.match(line):
                first_half = re.split(r'\s{2,}', self.file[num+2].strip())
                k_eff['first half k_eff'] = float(first_half[1])
                k_eff['first half stdev'] = float(first_half[2])

                second_half = re.split(r'\s{2,}', self.file[num+3].strip())
                k_eff['second half k_eff'] = float(second_half[1])
                k_eff['second half stdev'] = float(second_half[2])

                final_result = re.split(r'\s{2,}', self.file[num+4].strip())
                k_eff['final k_eff'] = float(final_result[1])
                k_eff['final stdev'] = float(final_result[2])
        return k_eff

    def get_active_cycles(self):
        """For a crit case, find the number of active and inactive cycles

        Returns:
            dict: A dictionary with entries for active and inactive cycles
        """
        cycles = {}
        for line in self.file:
            if "the minimum estimated standard deviation for the col/abs/tl keff estimator occurs with" in line:
                cycles["inactive"] = int(line.split()[12])
                cycles["active"] = int(line.split()[16])
        return cycles

    def get_cell_data(self):
        """Loop through the mcnp output to find the section containing the cell data

        Returns:
            list: The section of the output containing the cell data
        """
        PATTERN_cells_start = re.compile(r'1cells')
        PATTERN_cells_ends = re.compile(r' total')
        cell_section = []
        for line in self.file:
            if PATTERN_cells_start.match(line):
                cell_start_line = self.file.index(line)
                for row in self.file[cell_start_line:]:
                    cell_section.append(row.strip('\n'))
                    if PATTERN_cells_ends.match(row):
                        break
                break
        return cell_section

    def get_particle_populations(self, particle):
        """Find and return the section of the output data concerning populations of a particular particle.
        This function is in the EddyCase class rather than the Cell class as this data is in a separate part
        of the MCNP output from the rest of the cell data.

        Args:
            particle (str): either 'photon', 'neutron' or 'electron'

        Returns:
            list: the lines from the output file concerning populations of that particle,
                    or None if no particle information is found
        """
        PATTERN_run_terminated = re.compile(r'^\+\s+\d\d/\d\d/\d\d(.+)')
        PATTERN_particle_populations = re.compile(fr'^1{particle}\s+activity\sin\seach\scell.+')
        PATTERN_end_populations = re.compile(r'\s+total.+')
        for n, line in enumerate(self.file):
            if PATTERN_run_terminated.match(line):
                terminated = n
                for m, new_line in enumerate(self.file[terminated:], start=terminated):
                    if PATTERN_particle_populations.match(new_line):
                        for p in range(m, len(self.file)):
                            if PATTERN_end_populations.match(self.file[p]):
                                particle_populations = self.file[m:p + 1]
                                return particle_populations
        return None

    def create_cells(self):
        """Take the cell data and create objects of the Cell class

        Returns:
            list: a list of all the Cell objects

        """
        cell_list = []
        header = (self.cell_data[2].split())
        PATTERN_cell = re.compile(r'\s+\d+\s.+')
        for line in self.cell_data[1:]:
            if PATTERN_cell.match(line):
                info = line.split()
                neutron_importance = self.get_particle_importance(header, info, 'neutron')
                photon_importance = self.get_particle_importance(header, info, 'photon')
                electron_importance = self.get_particle_importance(header, info, 'electron')
                cell_list.append(Cell(
                    cell_number=info[1],
                    material_number=info[2],
                    atom_density=info[3],
                    gram_density=info[4],
                    volume=info[5],
                    mass=info[6],
                    neutron_importance=neutron_importance,
                    photon_importance=photon_importance,
                    electron_importance=electron_importance,
                    )
                )
        return cell_list

    def find_mcnp_particle_data(self, particle):
        """Find the particle data in the mcnp output.

        Args:
            particle (str): either neutron, photon or electron

        Returns:
            list: The part of the MCNP output with the particle data
        """
        particle_data = None  # ensure we return None if data is not found
        PATTERN_run_terminated = re.compile(r'^\+\s+\d\d/\d\d/\d\d(.+)')
        PATTERN_particle = {
            'neutron': re.compile(r'\sneutron\screation\s+tracks.+'),
            'photon': re.compile(r'\sphoton\screation\s+tracks.+'),
            'electron': re.compile(r'\selectron\screation\s+tracks.+')
        }
        # The particle data tables are different lengths for the different particle types
        particle_data_lines = {'neutron': 28, 'photon': 33, 'electron': 24}
        # Find the particle data and save to a list
        for n, line in enumerate(self.file):
            if PATTERN_run_terminated.match(line):
                terminated = n
                for m, new_line in enumerate(self.file[terminated:], start=terminated):
                    if PATTERN_particle[particle].match(new_line):
                        particle_data = self.file[m:m + particle_data_lines[particle]]
                        break
        return particle_data

    def create_particle(self, particle, particle_data):
        """Get the relevant information from the particle data (by calling another method), then create a Particle object

        Args:
            particle (str): Either 'neutron', 'photon' or 'electron'
            particle_data (list): The part of the MCNP output with the particle data for this particle

        Returns:
            Particle: The new particle object
        """
        creation_data, loss_data, total_data = self.sort_mcnp_particle_data(particle, particle_data)
        return Particle(
            particle_type=particle,
            creation_data=creation_data,
            loss_data=loss_data,
            total_data=total_data,
        )

    def get_tallies(self):
        """Find the tally sections in the MCNP output.

        Returns:
            tuple: the tally_list, f_types, F2_tallies, F4_tallies, F5_tallies, F6_tallies variables
        """
        tally_list = []   # list of all tallies
        f_types = []
        F2_tallies = {'neutrons': [], 'photons': [], 'electrons': []}
        F4_tallies = {'neutrons': [], 'photons': [], 'electrons': []}
        F5_tallies = {'neutrons': [], 'photons': [], 'electrons': []}
        F6_tallies = {'neutrons': [], 'photons': [], 'electrons': [], 'Collision Heating': []}

        PATTERN_run_terminated = re.compile(r'^\+\s+\d\d/\d\d/\d\d(.+)')
        PATTERN_tally_start = re.compile(r'^\s*1tally\s+\d+\s+nps.+')
        PATTERN_tally_end = re.compile(r'(^\s*(1tally|1status).+)')
        for n, line in enumerate(self.file):
            if PATTERN_run_terminated.match(line):
                terminated = n
                for m, new_line in enumerate(self.file[terminated:], start=terminated):
                    if PATTERN_tally_start.match(new_line):
                        tally_start = m
                        for o, other_line in enumerate(self.file[tally_start + 1:], start=tally_start + 1):
                            if PATTERN_tally_end.match(other_line):
                                tally_end = o
                                break
                        try:
                            tally_data = self.file[tally_start:tally_end]
                        except NameError:
                            print("Eddy did not find the end of one of the tally data sections.")
                            raise
                        tally_type = tally_data[1].split()[2]
                        # Create tally object
                        if tally_type == "2":
                            new_tally = F2Tally(tally_data)
                            F2_tallies[new_tally.particles].append(new_tally)
                        elif tally_type == "4":
                            new_tally = F4Tally(tally_data)
                            F4_tallies[new_tally.particles].append(new_tally)
                        elif tally_type == "5":
                            new_tally = F5Tally(tally_data)
                            F5_tallies[new_tally.particles].append(new_tally)
                        elif tally_type == "6" or tally_type == "6+":
                            new_tally = F6Tally(tally_data)
                            F6_tallies[new_tally.particles].append(new_tally)
                        else:
                            raise Exception("This tally type is not recognised")
                        assert new_tally, "EddyCase.get_tallies failed to create a recognised tally"
                        tally_list.append(new_tally)
                        if new_tally.f_type not in f_types:
                            f_types.append(new_tally.f_type)
                break
        return tally_list, f_types, F2_tallies, F4_tallies, F5_tallies, F6_tallies

    @staticmethod
    def sort_mcnp_particle_data(particle, data):
        """Separate out the photon data into creation, loss, and total data,
        then split each by indices and put into a dictionary

        Args:
            data (list): The part of the MCNP output with the particle data for this particle
            particle (str): either neutron, photon or electron

        Returns:
            a tuple containing
                creation_dict: The particle creation data,
                loss_dict: The particle loss data,
                total_dict: The particle neutron data,
        """
        # TODO: Make this whole method clearer

        indices = [0, 18, 27, 41, 55]
        particle_data_table_range = {
            'photon': range(3, 27),
            'neutron': range(3, 22),
            'electron': range(3, 18),
        }
        particle_summary_range = {
            'photon': range(28, 32),
            'neutron': range(23, 27),
            'electron': range(19, 23),
        }
        creation_dict = {}
        loss_dict = {}

        # get particle numbers from table in mcnp output
        for line_num in particle_data_table_range[particle]:
            # Separate creation data
            creation = data[line_num][:55]
            # split the line into 4 columns
            creation = [creation[indices[i]:indices[i + 1]].strip() for i in range(len(indices) - 1)]
            creation_dict[creation[0]] = {'tracks': int(creation[1]),
                                          'weight': float(creation[2]),
                                          'energy': float(creation[3]),
                                          }
            # Separate loss data
            loss_line = data[line_num][64:]
            # for photons there are fewer loss values than creation values, so we check if the line has contents
            if loss_line.isspace() is False:
                # split the line into 4 columns
                loss = [loss_line[indices[i]:indices[i + 1]].strip() for i in range(len(indices) - 1)]
                loss_dict[loss[0]] = {'tracks': int(loss[1]),
                                      'weight': float(loss[2]),
                                      'energy': float(loss[3]),
                                      }

        # get summary data for particle
        total_dict = {}
        for line_num in particle_summary_range[particle]:
            total = data[line_num][:42].strip(), data[line_num][42:53].strip()
            total_dict[total[0]] = total[1]

        return creation_dict, loss_dict, total_dict

    @staticmethod
    def get_particle_importance(header, data, particle):
        """For a particular particle type, determine the importance of that particle.
        This function is called for each cell as part of its creation.

        Args:
            header (list): The particle population table header.
            data (list): The section of the mcnp output with the cell data.
            particle (str): Either 'neutron', 'photon' or 'electron'.

        Returns:
            str: The particle importance, or 0.

        """
        if particle in header:
            particle_importance = data[header.index(particle) + 6]
        else:
            particle_importance = '0'
        return particle_importance



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

"""This module contains all the code relating to SCALE mixtures."""
# import from standard library:
import re


class Mixture:
    """Each mixture in the SCALE case is represented by a Mixture object"""
    def __init__(self, data):
        self.data = data
        self.number = data[0].split()[2]
        self.density = data[0].split()[5]
        try:
            self.temperature = data[0].split()[8] + 'K'
        except IndexError:
            self.temperature = "Not given in output file"
        self.isotopes = {}
        # There seem to be two formats of mixture table; if multigroup energies are used the table has a
        # 'nucmix' column, if continuous energies are used it does not.
        if 'nucmix' in data[1].split():
            try:
                self.isotopes = self.get_isotope_data_multigroup_format()
            except ValueError:
                print(ValueError)
        else:
            try:
                self.isotopes = self.get_isotope_data_continuous_format()
            except ValueError:
                print(ValueError)


    def get_isotope_data_multigroup_format(self):
        """Create a dictionary with the relevant parts of the mixture table
        Args:
            self: The Mixture object
        Returns:
            isotopes (dict): a dictionary where each entry is a name:nuclide pair, where
                            nuclide is a dict containing the relevant data about that nuclide
        """
        isotopes = {}
        headings = self.data[1].split('   ') # need 2 spaces so 'wgt. frac' is not split
        headings = [x.strip() for x in headings if x]  # remove empty strings
        for line in self.data[2:]:
            if line == '\n' or line == '':
                continue
            else:
                nuclide_index = headings.index("nuclide")
                nuclide = int(line.split()[nuclide_index])
                isotopes[nuclide] = {}
                isotopes[nuclide]['nuclide'] = nuclide
                isotopes[nuclide]['atom-density'] = float(line.split()[headings.index('atom-dens.')])
                isotopes[nuclide]['weight fraction'] = float(line.split()[headings.index('wgt. frac.')])
                isotopes[nuclide]['z-number'] = int(line.split()[headings.index('za')])
                isotopes[nuclide]['atomic weight'] = float(line.split()[headings.index('awt')])
                isotopes[nuclide]['title'] = line.split()[headings.index('nuclide title')].capitalize()
                isotopes[nuclide]['temperature'] = float(line.split()[headings.index('temp')])
        return isotopes


    def get_isotope_data_continuous_format(self):
        """Create a dictionary with the relevant parts of the mixture table
        Args:
            self: The Mixture object
        Returns:
            isotopes (dict): a dictionary where each entry is a name:nuclide pair, where
                            nuclide is a dict containing the relevant data about that nuclide
        """
        isotopes = {}
        headings = self.data[1].split('   ') # need 2 spaces so 'wgt. frac' is not split
        headings = [x.strip() for x in headings if x]  # remove empty strings
        for line in self.data[2:]:
            if line == '\n' or line == '':
                continue
            else:
                nuclide_index = headings.index("nuclide")
                nuclide = int(line.split()[nuclide_index])
                isotopes[nuclide] = {}
                isotopes[nuclide]['nuclide'] = nuclide
                isotopes[nuclide]['atom-density'] = float(line.split()[headings.index('atom-dens.')])
                isotopes[nuclide]['weight fraction'] = float(line.split()[headings.index('wgt. frac.')])
                isotopes[nuclide]['z-number'] = int(line.split()[headings.index('za')])
                isotopes[nuclide]['atomic weight'] = float(line.split()[headings.index('awt')])
                isotopes[nuclide]['title'] = line.split()[headings.index('title')].capitalize()
                isotopes[nuclide]['temperature'] = float(line.split()[headings.index('temp')])
        return isotopes


def get_mixture_data(output_data):
    """Get the part of the SCALE output concerning mixtures

    Args:
        output_data (list): The SCALE output data

    Returns:
        The lines from output_data concerning mixtures
    """
    pattern_mix = re.compile(r'^\s*mixing table\s*')
    for n, line in enumerate(output_data):
        if pattern_mix.match(line):
            for m, other_line in enumerate(output_data[n+1:], start=n+1):
                if "Cross section" in other_line or "*****" in other_line:
                    return output_data[n:m+1]


def create_mixtures(mix_data):
    """Create Mixture objects from the SCALE output data

    Args:
        mix_data (list): The lines from output_data concerning mixtures

    Returns:
        None, but creates Mixture objects
    """
    for n, line in enumerate(mix_data):
        if "mixture = " in line:
            for m, other_line in enumerate(mix_data[n+1:], start=n+1):
                if "mixture" in other_line or "Cross section" in other_line or "*****" in other_line:
                    mix = mix_data[n:m]
                    Mixture(mix)
                    break




import setuptools

with open("README.md", "r") as readme:
    long_description = readme.read()

setuptools.setup(
    name="eddy-mc-core",
    version="0.4.0",
    author="Cerberus Nuclear",
    author_email="nuclear@cerberusnuclear.com",
    description="Eddy, the MCNP and SCALE HTML output converter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Cerberus-Nuclear/Eddy-Core",
    packages=['eddymc_core', 'eddymc_core//mcnp', 'eddymc_core//scale', 'eddymc_core//static'],
    package_data={'eddymc_core//static': ['*']},
    install_requires=['Jinja2', 'importlib-resources'],
    classifiers=[
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



#!/usr/bin/python3
# Peter Evans
# Cerberus Nuclear Ltd

"""
This module holds the code that writes HTML to the html file.
"""


# Imports from standard library
import datetime
try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources
# Third party imports:
from jinja2 import Template, escape
# local imports
try:
    from .. import static
except ValueError:
    import static



def get_css():
    """Read in css from static css file

    Returns:
        inline_css (str): the whole static css file
    """
    inline_css = pkg_resources.read_text(static, 'style.css')
    return inline_css


def get_html(case):
    """Get the HTML from the jinja template

    Args:
        case (EddyMCNPCase): The EddyMCNPCase object for this Eddy run

    Returns:
        html (str): The completed html output
    """
    inline_css = get_css()

    html_template = pkg_resources.read_text(static, 'MCNP_template.html')
    template = Template(html_template)
    html = template.render(
        filename=case.filepath,
        case_name=case.name,
        # logo=f"{os.getcwd()}/static/logo.png",
        inline_css=inline_css,
        rundate=case.rundate,
        runtime=case.runtime,
        date=datetime.datetime.now().strftime("%Y/%m/%d"),
        time=datetime.datetime.now().strftime("%H:%M:%S"),
        scaling_factor=case.scaling_factor,
        ctme=case.ctme,
        nps=case.nps,
        parameters=case.parameters,
        fatal_errors=case.fatal_errors,
        lost_particles=case.lost_particles,
        k_eff=case.k_effective,
        tally_list=case.tally_list,
        tally_types=case.f_types,
        f2_tallies=case.F2_tallies,
        f4_tallies=case.F4_tallies,
        f5_tallies=case.F5_tallies,
        f6_tallies=case.F6_tallies,
        warnings=case.warnings,
        comments=case.comments,
        duplicate_surfaces=case.duplicate_surfaces,
        cell_list=case.cell_list,
        particle_list=case.particle_list,
        input_deck=case.mcnp_input,
        cycles=case.cycles,
        )
    return html



#!/usr/bin/env python3
# Peter Evans
# Cerberus Nuclear Ltd

"""This module contains the EddySCALECase class, which holds all the data relevant to a particular SCALE run.

"""

import re
from .scale_tallies import Tally
from .mixtures import Mixture


class EddySCALECase:
    def __init__(self, filepath, file, scaling_factor=1):
        self.filepath = filepath
        self.name = filepath.replace('\\', '/').split('/')[-1]
        self.file = file
        self.scaling_factor = scaling_factor

        # parse output
        self.scale_input = self.get_input()
        self.rundate, self.runtime = self.get_runtime()

        # Tallies
        self.tally_data = self.get_tally_data()
        self.tally_list = self.create_tallies()
        self.scale_tally_results()

        # Mixtures
        mixture_data = self.get_mixture_data()
        self.mixture_list = self.create_mixtures(mixture_data)

    def __repr__(self):
        return (f"This Eddy SCALE case considers file {self.name}\n"
                f"from {self.filepath}\n"
                f"The scaling factor is {self.scaling_factor}\n")

    def get_input(self):
        """Get the SCALE input deck from the output file

        Returns:
            list: The lines from the output containing the input data
        """
        for n, line in enumerate(self.file):
            if "Input Data:" in line:
                for m, other_line in enumerate(self.file[n + 1:], start=n + 1):
                    if other_line.lower() == 'end\n':
                        input_data = self.file[n + 2:m + 1]
                        return input_data

    def get_runtime(self):
        """Find the date and time that the SCALE case was run

        Returns:
            A tuple containing
                rundate (str): the date the case was run
                runtime (str): the time the case was run
        """
        for line in self.file:
            if "Job started on" in line:
                rundate = line.split()[6]
                # rearrange date from m/d/y to y/m/d
                y = rundate.split('/')[2]
                m = rundate.split('/')[1]
                d = rundate.split('/')[0]
                rundate = f"{y}/{m}/{d}"
                # time is fine as h/m/s
                runtime = line.split()[7]
                return rundate, runtime

    def get_tally_data(self):
        """Find the parts of the SCALE output with the tally data

        Returns:
            list: The lines from the SCALE output containing the tally data
        """
        for n, line in enumerate(self.file):
            if "Final Tally Results Summary" in line:
                for m, other_line in enumerate(self.file[n + 1:], start=n + 1):
                    if "Total Monaco cpu time for this problem" in other_line:
                        tally_data = self.file[n:m]
                        return tally_data
        # If the function has not returned the data, there is a problem
        raise Exception("Tally data incomplete.")

    def create_tallies(self):
        """Create Tally objects using the output data

        Returns:
            list: A list of Tally objects
        """
        tally_list = []
        for n, line in enumerate(self.tally_data):
            if "Region Tally" in line or "Point Detector" in line:
                tally_info = self.tally_data[n:n + 7]
                tally_list.append(Tally(tally_info))
        return tally_list

    def scale_tally_results(self):
        """Scale the results of each tally by the scaling factor"""
        if self.scaling_factor != 1:
            for tally in self.tally_list:
                tally.scale_results(self.scaling_factor)

    def get_mixture_data(self):
        """Get the part of the SCALE output concerning mixtures

        Returns:
            The lines from output_data concerning mixtures
        """
        pattern_mix = re.compile(r'^\s*mixing table\s*')
        for n, line in enumerate(self.file):
            if pattern_mix.match(line):
                for m, other_line in enumerate(self.file[n + 1:], start=n + 1):
                    if "Cross section" in other_line or "*****" in other_line:
                        return self.file[n:m]
        return None

    @staticmethod
    def create_mixtures(mix_data):
        """Create Mixture objects from the SCALE output data

        Args:
            mix_data (list): The lines from output_data concerning mixtures

        Returns:
            list: A list of Mixture objects
        """
        mixture_list = []
        for n, line in enumerate(mix_data):
            if "mixture = " in line:
                for m, other_line in enumerate(mix_data[n+1:], start=n+1):
                    if "mixture" in other_line or "Cross section" in other_line or "*****" in other_line:
                        mix = mix_data[n:m]
                        mixture_list.append(Mixture(mix))
                        break
        return mixture_list
